from __future__ import annotations

from typing import Any, Callable, Coroutine, Dict, List

import numpy as np

from .game_info import GameInfo
from .http_server import HttpServer
from .logger import Logger
from .message import Message
from .network_server_interface import INetworkServer
from .robot_status import RobotStatus
from .robot_control import RobotControl


class Server:
    """The MosHumanoid server."""

    class ClientInfo:
        """The information of the client."""

        def __init__(self, team: str, token: str):
            """Initializes the information of the client.

            Args:
                team: The name of the team.
                token: The token of the client.
            """

            self.team: str = team
            self.token: str = token

    _logger = Logger("Server")

    def __init__(self, port_controller: int, port_streaming: int, client_team_map: Dict[str, str]):
        """Initializes the server.

        Args:
            port_controller: The port of the controller server.
            port_streaming: The port of the streaming server.
            client_team_map: The map of the client and the team.
        """

        self._client_team_map: Dict[str, str] = client_team_map
        self._is_callback_registered: bool = False
        self._robot_control_callback_list: List[Callable[[str, RobotControl], Coroutine[Any, Any, None]]] = []

        # Components
        self._controller_network_server: INetworkServer = HttpServer(
            port_controller, list(client_team_map.keys()))
        self._streaming_network_server: INetworkServer = HttpServer(
            port_streaming, list(client_team_map.keys()))
        
        # Game data
        self._game_info: GameInfo | None = None

    async def start(self) -> None:
        """Starts the game."""

        if not self._is_callback_registered:
            await self._controller_network_server.register_callback(self._controller_callback)
            self._is_callback_registered = True

        await self._controller_network_server.start()
        await self._streaming_network_server.start()

    async def stop(self) -> None:
        """Stops the game."""

        await self._controller_network_server.stop()
        await self._streaming_network_server.stop()

    async def get_game_info(self) -> GameInfo | None:
        """Gets the information of the game.

        Returns:
            The information of the game.
        """

        return self._game_info
    
    async def set_game_info(self, game_info: GameInfo | None):
        """Sets the information of the game.

        Args:
            game_info: The information of the game.
        """

        self._game_info = game_info

    async def push_captured_image(self, token: str, image: np.ndarray) -> None:
        """Pushes the captured image to the client.

        Args:
            token: The token of the client.
            image: The captured image.
        """

        await self._streaming_network_server.send(Message({
            'type': 'push_captured_image',
            'bound_to': 'client',
            'data': image.tobytes(),
            'shape': list(image.shape),
        }), token)

    async def push_robot_status(self, token: str, robot_status: RobotStatus) -> None:
        """Pushes the status of the robot to the client.

        Args:
            token: The token of the client.
            robot_status: The status of the robot.
        """

        await self._controller_network_server.send(Message({
            'type': 'push_robot_status',
            'bound_to': 'client',
            'head': {
                'head_angle': robot_status.head_angle,
                'neck_angle': robot_status.neck_angle
            },
            'imu': {
                'acceleration': {
                    'x': robot_status.acceleration[0],
                    'y': robot_status.acceleration[1],
                    'z': robot_status.acceleration[2]
                },
                'angular_velocity': {
                    'pitch': robot_status.angular_velocity[0],
                    'yaw': robot_status.angular_velocity[1],
                    'roll': robot_status.angular_velocity[2]
                },
                'attitude_angle': {
                    'pitch': robot_status.attitude_angle[0],
                    'yaw': robot_status.attitude_angle[1],
                    'roll': robot_status.attitude_angle[2]
                }
            },
            'team': robot_status.team
        }), token)

    async def register_robot_control_callback(self, callback: Callable[[str, RobotControl], Coroutine[Any, Any, None]]) -> None:
        """Registers a callback for the robot control.

        Args:
            callback: The callback.
        """

        self._robot_control_callback_list.append(callback)

    async def _controller_callback(self, client_token: str, message: Message) -> None:
        try:
            message_bound_to: str = message.get_bound_to()

            # Filter out the messages bound to the client
            if message_bound_to == 'client':
                return

            message_type = message.get_type()

            if message_type == 'get_game_info':
                if self._game_info is None:
                    raise Exception("The game information is not ready.")

                if self._client_team_map.get(client_token, None) is None or \
                        self._game_info.score.get(self._client_team_map[client_token], None) is None:
                    raise Exception("The client is not in the game.")

                await self._controller_network_server.send(Message({
                    'type': 'get_game_info',
                    'bound_to': 'client',
                    'stage': self._game_info.stage.value,
                    'start_time': self._game_info.start_time.timestamp(),
                    'end_time': self._game_info.end_time.timestamp(),
                    'score': self._game_info.score,
                    'simulation_rate': self._game_info.simulation_rate
                }), client_token)

            elif message_type == 'push_robot_control':
                obj = message.to_dict()

                head = None
                movement = None
                kick = None

                if obj.get('head', None) is not None:
                    head = RobotControl.Head(
                        head_angle=obj['head'].get('head_angle', None),
                        neck_angle=obj['head'].get('neck_angle', None)
                    )

                if obj.get('movement', None) is not None:
                    movement = RobotControl.Movement(
                        x=obj['movement'].get('x', None),
                        y=obj['movement'].get('y', None),
                        omega_z=obj['movement'].get('omega_z', None)
                    )

                if obj.get('kick', None) is not None:
                    kick = RobotControl.Kick(
                        x=obj['kick']['x'],
                        y=obj['kick']['y'],
                        z=obj['kick']['z'],
                        speed=obj['kick']['speed'],
                        delay=obj['kick']['delay']
                    )

                for callback in self._robot_control_callback_list:
                    await callback(client_token, RobotControl(head, movement, kick))

        except Exception as e:
            self._logger.error(f"Failed to handle message: {e}")



from abc import ABC, abstractmethod
from typing import Any, Callable, Coroutine

from .message import Message


class INetworkServer(ABC):
    """Abstract interface for a network server."""

    @abstractmethod
    async def broadcast(self, msg: Message) -> None:
        """Broadcasts a message to all clients.

        Args:
            msg: The message to broadcast.
        """

        raise NotImplementedError

    @abstractmethod
    async def register_callback(self, callback: Callable[[str, Message], Coroutine[Any, Any, None]]) -> None:
        """Registers a callback function to be called when a message is received.

        Args:
            callback: The callback function to register. The arguments are the token of the client and the message.
        """

        raise NotImplementedError

    @abstractmethod
    async def send(self, msg: Message, token: str) -> None:
        """Sends a message to a client.

        Args:
            msg: The message to send.
            token: The token of the client to send the message to.
        """

        raise NotImplementedError

    @abstractmethod
    async def start(self) -> None:
        """Starts the server."""

        raise NotImplementedError

    @abstractmethod
    async def stop(self) -> None:
        """Stops the server."""

        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import Any, Callable, Coroutine

from .message import Message


class INetworkClient(ABC):
    """Abstract interface for a network client."""

    @abstractmethod
    async def connect(self) -> None:
        """Connects to the server."""

        raise NotImplementedError

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnects from the server."""

        raise NotImplementedError

    @abstractmethod
    async def register_callback(self, callback: Callable[[Message], Coroutine[Any, Any, None]]) -> None:
        """Registers a callback function to be called when a message is received.

        Args:
            callback: The callback function to register.
        """

        raise NotImplementedError

    @abstractmethod
    async def send(self, msg: Message) -> None:
        """Sends a message to the server.

        Args:
            msg: The message to send.
        """

        raise NotImplementedError



import datetime
from typing import Dict

from .game_stage_kind import GameStageKind


class GameInfo:
    """The information of the game."""

    def __init__(self, stage: GameStageKind, start_time: datetime.datetime,
                 end_time: datetime.datetime, score: Dict[str, int], simulation_rate: float):
        """Initializes the information of the game.

        Args:
            stage: The stage of the game.
            start_time: The start time of the game.
            end_time: The end time of the game.
            score: The score of the teams.
            simulation_rate: The simulation rate of the game.
        """

        self.stage: GameStageKind = stage
        self.start_time: datetime.datetime = start_time
        self.end_time: datetime.datetime = end_time
        self.score: Dict[str, int] = score
        self.simulation_rate: float = simulation_rate



from __future__ import annotations

from typing import Any, Dict

import bson
import jsonschema


class Message:
    """The message to communicate with servers."""

    _SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "type": {
                "type": "string"
            },
            "bound_to": {
                "type": "string",
                "enum": [
                    "client",
                    "server"
                ]
            }
        },
        "required": [
            "type",
            "bound_to"
        ]
    }

    def __init__(self, payload: Dict[str, Any] | bytes):
        """Initializes the message.

        Args:
            payload: The payload of the message.
        """

        self._payload: Dict[str, Any] = {}

        if isinstance(payload, bytes):
            self._payload = bson.decode(payload)
        elif isinstance(payload, dict):
            self._payload = payload
        else:
            raise TypeError("The payload must be a dictionary or bytes.")

        validator = jsonschema.Draft7Validator(Message._SCHEMA)
        if not validator.is_valid(self._payload):
            raise ValueError("The JSON is not valid.")

    def __str__(self) -> str:
        """Converts the message to a string.

        Returns:
            The string of the message.
        """

        return str(self._payload)

    def to_bytes(self) -> bytes:
        """Converts the message to BSON bytes.

        Returns:
            The BSON bytes of the message.
        """

        return bson.encode(self._payload)

    def to_dict(self) -> Dict[str, Any]:
        """Converts the message to a dictionary.

        Returns:
            The dictionary of the message.
        """

        return self._payload

    def get_bound_to(self) -> str:
        """Gets the direction of the message.

        Returns:
            The direction of the message. "server" if the message is sent to the server, 
            "client" if the message is sent to the client.
        """

        return self._payload["bound_to"]

    def get_type(self) -> str:
        """Gets the type of the message.

        Returns:
            The type of the message.
        """

        return self._payload["type"]



import asyncio
from typing import Any, Callable, Coroutine, List

import aiohttp

from .logger import Logger
from .message import Message
from .network_client_interface import INetworkClient


class HttpClient(INetworkClient):
    """The HTTP client to communicate with the server."""

    _FETCH_INTERVAL = 0.05

    _logger = Logger("HttpClient")

    def __init__(self, host: str, port: int, token: str):
        """Initializes a new instance of the HttpClient class.

        Args:
            host: The server address.
            port: The server port.
            token: The token of the client.
        """

        self._url: str = f"http://{host}:{port}/"
        self._token: str = token

        self._callback_list: List[Callable[[Message],
                                           Coroutine[Any, Any, None]]] = []
        self._task_list: List[asyncio.Task] = []

        self._session = aiohttp.ClientSession()

    async def connect(self) -> None:
        """Connects to the server."""

        self._task_list.append(
            asyncio.create_task(self._loop())
        )

    async def disconnect(self) -> None:
        """Disconnects from the server."""
        
        for task in self._task_list:
            task.cancel()

        await self._session.close()

    async def register_callback(self, callback: Callable[[Message], Coroutine[Any, Any, None]]) -> None:
        """Registers a callback function to be called when a message is received.

        Args:
            callback: The callback function to register.
        """

        self._callback_list.append(callback)

    async def send(self, msg: Message) -> None:
        """Sends a message to the server.

        Args:
            msg: The message to send.
        """
        
        try:
            async with self._session.post(
                self._url,
                data=msg.to_bytes(),
                headers={"Authorization": f"Bearer {self._token}"}
            ) as response:
                if response.status == 204:
                    return

                else:
                    raise Exception(
                        f"HTTP status: {response.status} {response.reason}"
                    )

        except Exception as e:
            self._logger.error(f"Failed to send: {e}")

    async def _loop(self) -> None:
        while True:
            await asyncio.sleep(HttpClient._FETCH_INTERVAL)

            try:
                async with self._session.get(
                    self._url,
                    headers={"Authorization": f"Bearer {self._token}"}
                ) as response:
                    if response.status == 204:
                        continue

                    elif response.status == 200:
                        msg = Message(await response.read())
                        for callback in self._callback_list:
                            await callback(msg)

                    else:
                        raise Exception(
                            f"HTTP status: {response.status} {response.reason}"
                        )

            except Exception as e:
                self._logger.error(f"Failed to receive: {e}")
                await asyncio.sleep(1)
                continue



import sys
import threading
from datetime import datetime

from termcolor import cprint


class Logger:
    _FORMAT = ""

    _lock = threading.Lock()

    def __init__(self, namespace: str) -> None:
        super().__init__()

        gettrace = getattr(sys, "gettrace", None)
        if gettrace is not None and gettrace():
            self._debug = True
        else:
            self._debug = False

        self._namespace = namespace

    def debug(self, message: str) -> None:
        if not self._debug:
            return

        with Logger._lock:
            cprint(f'{Logger._get_current_time_string()} ',
                   color='cyan', end='')
            cprint(f'DEBUG ', color='dark_grey', end='')
            cprint(f'[{self._namespace}] {message}', color='dark_grey', end='')
            print()

    def info(self, message: str) -> None:
        with Logger._lock:
            cprint(f'{Logger._get_current_time_string()} ',
                   color='cyan', end='')
            cprint(f'INFO  ', color='blue', end='')
            cprint(f'[{self._namespace}] {message}', color='white', end='')
            print()

    def warn(self, message: str) -> None:
        with Logger._lock:
            cprint(f'{Logger._get_current_time_string()} ',
                   color='cyan', end='')
            cprint(f'WARN  ', color='yellow', end='')
            cprint(f'[{self._namespace}] {message}', color='yellow', end='')
            print()

    def error(self, message: str) -> None:
        with Logger._lock:
            cprint(f'{Logger._get_current_time_string()} ',
                   color='cyan', end='')
            cprint(f'ERROR ', color='red', end='')
            cprint(f'[{self._namespace}] {message}', color='red', end='')
            print()

    @staticmethod
    def _get_current_time_string() -> str:
        return datetime.now().strftime(f'%H:%M:%S')



from enum import Enum


class GameStageKind(Enum):
    """Game stage kind.
    
    Attributes:
        READY: The game is ready to start.
        IN_PROGRESS: The game is in progress.
        PAUSED: The game is paused.
        FINISHED: The game is finished.
        ABORTED: The game is aborted.
    """

    READY = "ready"
    IN_PROGRESS = "in_progress"
    PAUSED = "paused"
    FINISHED = "finished"
    ABORTED = "aborted"



from .client import Client
from .game_info import GameInfo
from .game_stage_kind import GameStageKind
from .http_client import HttpClient
from .http_server import HttpServer
from .message import Message
from .network_client_interface import INetworkClient
from .network_server_interface import INetworkServer
from .robot_control import RobotControl
from .robot_status import RobotStatus
from .server import Server

__all__ = [
    "Client",
    "GameInfo",
    "GameStageKind",
    "HttpClient",
    "HttpServer",
    "Message",
    "INetworkClient",
    "INetworkServer",
    "RobotControl",
    "RobotStatus",
    "Server"
]



import asyncio
import re
from typing import Any, Callable, Coroutine, Dict, List

from aiohttp import web

from soccerxcomm.message import Message

from .logger import Logger
from .network_server_interface import INetworkServer


class HttpServer(INetworkServer):
    """The HTTP server to communicate with the client."""

    _MESSAGE_QUEUE_MAX_SIZE = 10

    _logger = Logger("HttpServer")

    def __init__(self, port: int, token_list: List[str]):
        """Initializes a new instance of the HttpServer class.

        Args:
            port: The port of the server.
            token_list: The list of tokens that are allowed to connect to the server.
        """
        self._port: int = port

        self._callback_list: List[Callable[[
            str, Message], Coroutine[Any, Any, None]]] = []
        self._message_queue_dict: Dict[str, asyncio.Queue[Message]] = {
            token: asyncio.Queue(HttpServer._MESSAGE_QUEUE_MAX_SIZE) for token in token_list
        }

        # Initialize the HTTP server.
        self._app = web.Application()
        self._app.add_routes([
            web.get('/', self._on_get),
            web.post('/', self._on_post)
        ])
        self._runner = web.AppRunner(self._app)

    async def broadcast(self, msg: Message) -> None:
        """Broadcasts a message to all clients.

        Args:
            msg: The message to broadcast.
        """

        for token in self._message_queue_dict.keys():
            await self.send(msg, token)

    async def register_callback(self, callback: Callable[[str, Message], Coroutine[Any, Any, None]]) -> None:
        """Registers a callback function to be called when a message is received.

        Args:
            callback: The callback function to register. The arguments are the token of the client and the message.
        """

        self._callback_list.append(callback)

    async def send(self, msg: Message, token: str) -> None:
        """Sends a message to a client.

        Args:
            msg: The message to send.
            token: The token of the client to send the message to.
        """

        if token not in self._message_queue_dict:
            raise Exception("The token is not registered.")
        
        message_queue: asyncio.Queue[Message] = self._message_queue_dict[token]
        
        if message_queue.full():
            # Remove the oldest message.
            message_queue.get_nowait()
        
        await self._message_queue_dict[token].put(msg)

    async def start(self) -> None:
        """Starts the server."""

        await self._runner.setup()
        site = web.TCPSite(self._runner, port=self._port)
        await site.start()

    async def stop(self) -> None:
        """Stops the server."""
        
        await self._runner.cleanup()

    async def _on_get(self, request: web.Request) -> web.Response:
        try:
            auth_header = request.headers.get('Authorization')
            if auth_header is None:
                return web.Response(status=401)

            match_result = re.search(r'Bearer (\w*)', auth_header)
            if match_result is None:
                return web.Response(status=401)

            token = match_result.group(1)

            if token not in self._message_queue_dict:
                return web.Response(status=403)

            if self._message_queue_dict[token].empty():
                return web.Response(status=204)

            msg: Message = await self._message_queue_dict[token].get()
            self._logger.debug(f"Sending message: {msg}")
            return web.Response(body=msg.to_bytes())

        except Exception as e:
            self._logger.error(
                f"Failed to handle GET from {request.remote}: {e}")
            return web.Response(status=400)

    async def _on_post(self, request: web.Request) -> web.Response:
        try:
            auth_header = request.headers.get('Authorization')
            if auth_header is None:
                return web.Response(status=401)

            match_result = re.search(r'Bearer (\w*)', auth_header)
            if match_result is None:
                return web.Response(status=401)

            token = match_result.group(1)
            if token not in self._message_queue_dict:
                return web.Response(status=403)

            msg = Message(await request.read())
            self._logger.debug(f"Received message: {msg}")
            for callback in self._callback_list:
                await callback(token, msg)

            return web.Response(status=204)

        except Exception as e:
            self._logger.error(
                f"Failed to handle POST from {request.remote}: {e}")
            return web.Response(status=400)



from __future__ import annotations


class RobotControl:
    """Robot control commands."""

    class Head:
        """Head control commands."""

        def __init__(self, head_angle: float | None = None, neck_angle: float | None = None):
            """Initializes the head control commands.

            Args:
                head_angle: The head angle in degrees.
                neck_angle: The neck angle in degrees.
            """

            self.head_angle: float | None = head_angle
            self.neck_angle: float | None = neck_angle

    class Movement:
        """Movement control commands."""

        def __init__(self, x: float | None = None, y: float | None = None, omega_z: float | None = None):
            """Initializes the movement control commands.

            Args:
                x: The x-axis velocity.
                y: The y-axis velocity.
                omega_z: The angular velocity around the z-axis.
            """

            self.x: float | None = x
            self.y: float | None = y
            self.omega_z: float | None = omega_z

    class Kick:
        """Kick control commands."""

        def __init__(self, x: float, y: float, z: float, speed: float, delay: float):
            """Initializes the kick control commands.

            Args:
                x: The x-axis velocity.
                y: The y-axis velocity.
                z: The z-axis velocity.
                speed: The speed of the kick.
                delay: The delay of the kick.
            """

            self.x: float = x
            self.y: float = y
            self.z: float = z
            self.speed: float = speed
            self.delay: float = delay

    def __init__(self, head: Head | None = None, movement: Movement | None = None, kick: Kick | None = None):
        """Initializes the robot control commands.

        Args:
            head: The head control commands.
            movement: The movement control commands.
            kick: The kick control commands.
        """

        self.head: RobotControl.Head | None = head
        self.movement: RobotControl.Movement | None = movement
        self.kick: RobotControl.Kick | None = kick



from __future__ import annotations

import asyncio
import datetime
from typing import Any, Dict, List

import numpy as np

from .game_info import GameInfo
from .game_stage_kind import GameStageKind
from .http_client import HttpClient
from .logger import Logger
from .message import Message
from .network_client_interface import INetworkClient
from .robot_control import RobotControl
from .robot_status import RobotStatus


class Client:
    """The client to send commands to the server."""

    _logger = Logger("Client")

    def __init__(self, host: str, port_controller: int, port_streaming: int, token: str):
        """Initializes the client.

        Args:
            host: The server address.
            port_controller: The port of the controller server.
            port_streaming: The port of the streaming server.
            token: The token of the game.
        """

        self._is_callback_registered: bool = False
        self._controller_network_client: INetworkClient = HttpClient(
            host, port_controller, token)
        self._streaming_network_client: INetworkClient = HttpClient(
            host, port_streaming, token)
        self._task_list: List[asyncio.Task] = []

        self._game_info: GameInfo | None = None
        self._robot_status: RobotStatus | None = None

        self._captured_image: np.ndarray | None = None

    async def connect(self) -> None:
        """Connects to the server."""

        if not self._is_callback_registered:
            await self._controller_network_client.register_callback(self._controller_callback)
            await self._streaming_network_client.register_callback(self._streaming_callback)
            self._is_callback_registered = True

        await self._controller_network_client.connect()
        await self._streaming_network_client.connect()

        self._task_list.append(asyncio.create_task(self._controller_loop()))

    async def disconnect(self) -> None:
        """Disconnects from the server."""

        for task in self._task_list:
            task.cancel()

        self._task_list.clear()

        await self._controller_network_client.disconnect()
        await self._streaming_network_client.disconnect()

    async def get_captured_image(self) -> np.ndarray | None:
        """Gets the captured image.

        Returns:
            The captured image.
        """

        return self._captured_image

    async def get_game_info(self) -> GameInfo | None:
        """Gets the game information.

        Returns:
            The game information.
        """

        return self._game_info

    async def get_robot_status(self) -> RobotStatus | None:
        """Gets the robot status.

        Returns:
            The robot status.
        """

        return self._robot_status

    async def push_robot_control(self, robot_control: RobotControl) -> None:
        """Pushes the control of the robot to the server.

        Args:
            robot_control: The control of the robot.
        """

        obj: Dict[str, Any] = {
            'type': 'push_robot_control',
            'bound_to': 'server',
        }

        if robot_control.head is not None:
            obj['head'] = {}
            if robot_control.head.head_angle is not None:
                obj['head']['head_angle'] = robot_control.head.head_angle
            if robot_control.head.neck_angle is not None:
                obj['head']['neck_angle'] = robot_control.head.neck_angle

        if robot_control.movement is not None:
            obj['movement'] = {}
            if robot_control.movement.x is not None:
                obj['movement']['x'] = robot_control.movement.x
            if robot_control.movement.y is not None:
                obj['movement']['y'] = robot_control.movement.y
            if robot_control.movement.omega_z is not None:
                obj['movement']['omega_z'] = robot_control.movement.omega_z

        if robot_control.kick is not None:
            obj['kick'] = {}
            obj['kick']['x'] = robot_control.kick.x
            obj['kick']['y'] = robot_control.kick.y
            obj['kick']['z'] = robot_control.kick.z
            obj['kick']['speed'] = robot_control.kick.speed
            obj['kick']['delay'] = robot_control.kick.delay

        await self._controller_network_client.send(Message(obj))

    async def _controller_callback(self, msg: Message) -> None:
        try:
            message_bound_to: str = msg.get_bound_to()

            if message_bound_to == 'server':
                return

            message_type: str = msg.get_type()

            if message_type == 'get_game_info':
                self._game_info = GameInfo(
                    stage=GameStageKind(str(msg.to_dict()['stage'])),
                    start_time=datetime.datetime.fromtimestamp(
                        int(msg.to_dict()['start_time'])),
                    end_time=datetime.datetime.fromtimestamp(
                        int(msg.to_dict()['end_time'])),
                    score=msg.to_dict()['score'],
                    simulation_rate=float(msg.to_dict()['simulation_rate']),
                )

            elif message_type == 'push_robot_status':
                self._robot_status = RobotStatus(
                    head_angle=float(msg.to_dict()['head']['head_angle']),
                    neck_angle=float(msg.to_dict()['head']['neck_angle']),
                    acceleration=np.array([
                        float(msg.to_dict()['imu']['acceleration']['x']),
                        float(msg.to_dict()['imu']['acceleration']['y']),
                        float(msg.to_dict()['imu']['acceleration']['z']),
                    ]),
                    angular_velocity=np.array([
                        float(msg.to_dict()['imu']
                              ['angular_velocity']['pitch']),
                        float(msg.to_dict()['imu']['angular_velocity']['yaw']),
                        float(msg.to_dict()['imu']
                              ['angular_velocity']['roll']),
                    ]),
                    attitude_angle=np.array([
                        float(msg.to_dict()['imu']['attitude_angle']['pitch']),
                        float(msg.to_dict()['imu']['attitude_angle']['yaw']),
                        float(msg.to_dict()['imu']['attitude_angle']['roll']),
                    ]),
                    team=msg.to_dict()['team']
                )

        except Exception as e:
            self._logger.error(f'Failed to handle message: {e}')

    async def _controller_loop(self) -> None:
        while True:
            try:
                await asyncio.sleep(1)

                await self._controller_network_client.send(Message(
                    {
                        "type": "get_game_info",
                        "bound_to": "server"
                    }
                ))

            except Exception as e:
                self._logger.error(f'Failed to get info: {e}')

    async def _streaming_callback(self, msg: Message) -> None:
        try:
            message_bound_to: str = msg.get_bound_to()

            if message_bound_to == 'server':
                return

            message_type: str = msg.get_type()

            if message_type == 'push_captured_image':
                data: bytes = msg.to_dict()['data']
                shape: List[int] = msg.to_dict()['shape']
                self._captured_image = np.frombuffer(
                    data, dtype=np.uint8).reshape(shape)

        except Exception as e:
            self._logger.error(f'Failed to handle message: {e}')



import numpy as np


class RobotStatus:
    """The status of the robot."""

    def __init__(self, head_angle: float, neck_angle: float, acceleration: np.ndarray, angular_velocity: np.ndarray, attitude_angle: np.ndarray, team: str):
        """Initializes the status of the robot.

        Args:
            head_angle: The angle of the head.
            neck_angle: The angle of the neck.
            acceleration: The acceleration of the robot.
            angular_velocity: The angular velocity of the robot.
            attitude_angle: The attitude angle of the robot.
            team: The team of the robot.
        """

        # Validate the shapes
        if acceleration.shape != (3,):
            raise Exception("The shape of the acceleration must be (3,).")
        if angular_velocity.shape != (3,):
            raise Exception("The shape of the angular velocity must be (3,).")
        if attitude_angle.shape != (3,):
            raise Exception("The shape of the attitude angle must be (3,).")

        self.head_angle: float = head_angle
        self.neck_angle: float = neck_angle
        self.acceleration: np.ndarray = acceleration
        self.angular_velocity: np.ndarray = angular_velocity
        self.attitude_angle: np.ndarray = attitude_angle
        self.team: str = team
